; 
; Copyright 1999 by Konrad Rzeszutek
; Author: Konrad Rzeszutek 
; Date:   Sat 3rd Apr 1999
; File:   stones.lsp
;
; Notes:
; ======
;
;  This board game is based on the African game Kalah. This game serves to present
; the alpha-beta-prunning algorhytm and min-max algorhytm. For information regarding
; how the game should be played and as well examples, please contact Prof. Bill Grene
; (bill@cs.uno.edu).
;
; For convience the game also offers for simplicity three types of play:
;
;   [o] user vs. user.
;   [o] user vs. computer.
;   [o] computer vs. computer.
; 
; The game also offers tunning of the difficulty of the computer thinking.
;
; Developers Notes:
; =================
;
; This board game operates on a DFS tree that implements a large amount of nodes.
;
; Each node consists of:
;  * a board
;  * a bank
;  * depth of the node
;  * "goodness" of the node
;  * list of descdentents
;  * parent-node
;
; [board]:
;
; The board is a list representation of the board. It consists of 8 elements[pits] (can be
; changed by using *amount-of-pits*). The first four belong to the "BOTTOM" player, while
; the rest four to the "TOP" player. It has to be noted, that the "BOTTOM" and "TOP" names 
; are just used for convenience of imaging the game. The list is positioned in following
;  order on a board:
;
;    [7] [6] [5] [4]
;    [0] [1] [2] [3]
;
; As you can see, the bottom elements correspond to the first four element.
; If it still hasn't clicked, just run : 
;  (show-pits-game-message (make-node '(4 4 4 4 4 4 4 4) '(0 0) 0 0 nil nil))
;
; [bank]:
;
; The bank is a two-elemnt list. The first element belongs to the "BOTTOM" player, while
; the other belongs .... ..... ... to "TOP" player. You knew that, didn't you? Good.
;
; [depth/goodness]:
;
; The depth of the node/goodness of the node is only used by the computer. 
; Names of these two variables make their purpose  evident.
; Please note that the goodness of the variable as the list is build depends on the
; the depth. Why? Becuase otherwise the min/max algorhytm wouldn't work.
; Picture that we are building a small tree using DFS and that we just found the terminal
; node.
;
;   Depth: 
;    1                 A (root, since this maximizes, the goodness value is neg-infinity)
;                     / 
;    2              B (this minimizes, goodness value is pos-infinity)
;                  /
;    3            C  (this is our terminal node, its goodness value is equal to its static
;                     evaluaton function)
;
; So, from this example, we know that based on the depth we can
; decide whether a node is maximizing/minimizing or is a terminal node.
;
; Hint: In this program, odd maximes, even minimizes, while *lookahead-depth* defines
; on what depth the terminal nodes will be located.
; 
; Keep one thing in mind when altering the *lookahead-depth*. The nodes are built based
; on b^d equation. Branching factor is almost everytime 4, so for a depth 4, we get
; 4^4 nodes = 256. Now just think about 4^10 = 1,048,576 nodes. Takes helluva of time to
; create all of them (don't worry about space -- the algorhytm deletes proccessed
; nodes).
;
; Conclusion:
; ===========
;
; This prologue should be enough to understand how this game works. If there are still
; questions regarding the implementation just follow the comments in the functions. 
;
; *WARNING*: 
; They all are intended for a person knowing how alpha-beta-prunning, min/max and alike works.
;
; If you don't have a clue what this is about, then read this:
;
;  Essentials of Artificial Intelligence, by Matt Ginsbert, Morgan Kaufmann Publ.
;  ISBN 1-55860-221-6
;
; Well, start the game invoking the main module:
;
; (start-game)
;
; Comments
; ========
;
;  The source code and as well the log file is located in 
;   ftp://mars.cs.uno.edu/pub/projects/stone/stone.tgz
;


(defvar *type-of-players* '(0 0))  ; Two players, in this case both are human.
                                   ; 0 defines human, 1 defines computer
(defvar *amount-of-pits* 8)
(defvar *to-win* 17)               ; amount of stones req to win
(defvar *to-tie* 16)               ; amount of stones to tie (both parties)

(defvar *which-player* 0)          ; 0 is BOTTOM, 1 is TOP

(defvar *lookahead-depth* 5)       ; difficulty of the game (3-10). You can go higher, 
                                   ; but it is not _advisable_.
(defvar *neg-infinity* -666)       
(defvar *pos-infinity* 666)



; +==========================================================================+
; | Main game methods                                                        |
; |                                                                          |
; +==========================================================================+
;
(defun change-type-of-players-to (first-p second-p)
  
  ; exclusivly used to setup what kind of player we are playing with.

  (setf (nth 0 *type-of-players*) first-p)
  (setf (nth 1 *type-of-players*) second-p))

(defun start-game ()
 
  
  (let
      ((which-game) ; which option from the menu
       (game-messages '(lost-game-message tie-game-message  won-game-message))
                           ; all important messages dealing with game
       )
    (loop
      (menu-game-message)     
      (setq which-game (read)) ; read from user
      (cond 
       ( (numberp which-game)

	 (cond
	  ( (= which-game 0) ; Two human players
	    (change-type-of-players-to 0 0)
	    (funcall (nth (game) game-messages))) ; Call the function 'game' and based on 
					; its return variable, display appropiate message
					; from game-message
	  ( (= which-game 1) ; One human against the machine
	    (change-type-of-players-to 0 1)
	    (funcall (nth (game) game-messages)))
	  ( (= which-game 2) ; Two computer players
	    (change-type-of-players-to 1 1) ;
	    (funcall (nth (game) game-messages)))
	  ( (= which-game 3) ; Change computers "IQ"
	    (loop
	      (format t"~%Please input my the new difficulty level from 3-10 >")
	      (setq *lookahead-depth* (read))
	      (if (and (numberp *lookahead-depth*) (<= 3 *lookahead-depth* 10)) ; Don't try anything above 10!
		  (return)
		(format t "~%Wrong choice Home Sapiens sapiens. Please pick again!~%"))))
	  ( (= which-game 4) ; Quit
	    (return)))
	 )
       (t (format t "~%Are you nuts? I need numbers!~%")))
    )          
  )
)

;
; The computer move.
;
; Require:
;
;  node - node received from the opponent.
;  *which-player* has to be set to the computer side.
; 
(defun game-computer-move (node)
  
  ; Make a move based upon 'node'

  (setf (get node 'g-value) *neg-infinity*) ; We need to reset those.
  (setf (get node 'childrens) nil)
  (setf (get node 'depth) 1)
  (show-pits-game-message node) ; That's what the opponents choose.
  (turn-board-around)           ; This is needed for the min/max. Look in a-b-prunning
				; function for more info.
  
  (format t " Just a minute. I'm busy pondering.")
  (let
      ((children (a-b-prunning (list node)))) ; Getting a list of four computer nodes.
    
       (setq children (if (null children)  ; The computer might decide its better not 
					   ; to do anything.
			  (setq children (list node)) ; if so, then lets just return 
					; the node we got from opponent
			children)) ; but we came with our own answer, then this

       (cond ; those conditions are required for computer vs. computer and computer vs. player
	     ; The list of children is sorted based on their g-value and the first is picked for
             ; the computer move. Both computers want their g-value to be the
	     ; most smallest If the '>' is switched for '<' 
	     ; the computer would actually help the opponent! This is better then "god-mode"!! :p
	( (= *which-player* 0)
	  (setq children (sort children  #'(lambda (x y) 
					     (when (> (g-value-of x) (g-value-of y)) t)))))
	( (= *which-player* 1)
	  (setq children (sort children  #'(lambda (x y) 
					     (when (> (g-value-of x) (g-value-of y)) t)))))
	)       
       (first children) 
       ))

;
; The weak computer move
;
; Require:
;
; node - has to be from the opponent.
;
(defun game-human-move (node)

  ; Make a move.
  (let
      ((answer)) ; the choosen pit
    (loop
      (show-pits-game-message node) ; show the board
      
      (setq answer (read))
      (if (<= 1 answer 4)
	  (return) (format t "~%Wrong pit. Please pick again.~%"))
    )
    (format t "You are ~A ~%"  *which-player*)
    (setq answer (take-out-from-pit node answer)) ; Do the magic
    (turn-board-around) ; For the opponent.
    (if (null answer) ; In case he doesn't want to make his move
	node
      answer)    
    )
  )
  
;
; Just one game.
;
(defun game ()

  (setq *which-player* 1)  ; BOTTOM player always starts.
  (let
      ((players '(game-human-move game-computer-move)) ; This corresponds to
					;  *type-of-players* and defines whetever 
					; the opponent is human or computer.
       (node (make-node '(4 4 4 4 4 4 4 4) '(0 0) 1 *neg-infinity* nil nil)) 
					; the starting node.
       )
    (loop 
      (setq node (funcall (nth (nth *which-player* *type-of-players*) players) node)) 
    ; ^- Talk about complexity! The middle part defines whetever the current player
    ; is human(0) or computer(1). Then the first nth (from left), looks at the list players and 
    ; gets the corresponding game-xxx-move. funcall calls the method with the node.
    ; The whole function returns the new node which gets assigned to node.
    ; Easy, eh? :-k
      (when (>= (nth 0 (bank-of node)) *to-win*) ; first player won
	(return 2))
      (when (>= (nth 1 (bank-of node)) *to-win*)
	(return 2))
      (when (and (= (nth (current-player-starting-index-is) (board-of node)) 0) ; checks to see if we can move
		 (= (nth (+ (current-player-starting-index-is) 1) (board-of node)) 0) ; at all
		 (= (nth (+ (current-player-starting-index-is) 2) (board-of node)) 0)
		 (= (nth (+ (current-player-starting-index-is) 3) (board-of node)) 0))
	
	(when (and (= (nth 0 (bank-of node)) *to-tie*) (= (nth 1 (bank-of node)) *to-tie*)) ; did both parties tie?
	  (return 1)) ; yes, we tied
	(return 0)) ; no, somebody lost
      )
    )
  )

;
; Alpha-Beta prunning tree with min/max technique.
;
; Visualize a DFS tree where each node has its own "goodness" value. When the tree
; is created each node (except the terminal ones) has either a maximum or 
; minimum value. When the program gets to the terminal node, that node
; receives a value based on static evalutation function. This "goodness"
; value gets compared with the parent's value, and depending on the depth of the node
; the min or max function gets applied to the "goodness" of parent (Vp) and
; the "goodness" of the node (Vx). Hence, if the parent is maximizing,
; the goodness of the parent (Vp) will be set to the max  of (Vx, Vp), and if parent 
; is minimizing, the Vp  will be = min (Vx, Vp).
;
; This process will continue so long as all the nodes are created and have
; been proccessed. Hence, the value returned will be a list of the children
; of the the utermost first node.
;

(defun a-b-prunning (tree)

  (cond 
   ((null tree) ; When we are finish.
    nil)
   (t
    (let 
	((first-node nil) ; The first element from the tree
	 (Vx)             ; Goodness value of the 'first-node'
	 (Vp)             ; Goodness value of the 'first-node' parent
	 (parent-node nil)
	 (next-node)      ; Will be created
	 (move-list '(take-out-from-pit-1 take-out-from-pit-2 take-out-from-pit-3 take-out-from-pit-4)) ; All legal moves
	 )
      (setq first-node (pop tree)) ; Removes the first as well
      (setq Vx (g-value-of first-node))
      (setq parent-node (parent-of first-node))
      (setq Vp (g-value-of parent-node))

      (cond 
       ; ------
       ; Legitimate g-values
       ; ------
       ( (and (not (= *neg-infinity* Vx)) (not (= *pos-infinity* Vx))) 
	 ; does the first-node  have legitimate g-values?
	 (cond 
	  ( (null parent-node) ; Yes, and it is the root of the tree
	    (list first-node))
	  (t
	       (cond  
		      ( (oddp (depth-of parent-node)) ; parent is odd (maximizing)
			(setf (get parent-node 'g-value) (max Vx Vp))) 
                        ; set it to the maximum of parent-node and this current-one
		      ( (evenp (depth-of  parent-node)) ; parent is even (minimizing)
			(setf (get parent-node 'g-value) (min Vx Vp))) 
                        ; set to minimum of both values.
		      )
	       tree))) ; return the rest for processing. Note that this current element
                       ; -- "first-node" is not on the list

       ; ------
       ; Illegimate depth -- not on terminal nodes
       ; -------
       ( (not (= (depth-of first-node) *lookahead-depth*)) 
         ; is it at the correct depth?

	 (push first-node tree) ; putting the element back on so that children can access
                                ; the parent 
	 (turn-board-around)
	 (dolist (move move-list)
	   (setq next-node (funcall move first-node)) ; Call each of the four 
           ; take-out-from-pit-x and put the resulting node in 'next-node'
	   (when next-node
	     (push next-node tree)
	     (setf (get first-node 'childrens) (nconc (list next-node) (children-of first-node)))
					; ^- adds the new node to the tree and to the 
                                        ; parent's children list.
	     (setq tree (a-b-prunning tree)) ; process the child.
	     ; we are finish with the processing the child, so lets do some prunning:
	     (cond 
	      ( (= (depth-of first-node) 1) ; Is this the root level?
		nil)
	      ( t ; no, some other level
		(cond 
		 ((oddp (depth-of parent-node)) ; is parent maximizing?
                  ; this is checking to see if prunning has to be done.
		  (if (< Vx Vp) ; Since our child updated "our" Vx value, we compare
		      ; it to "our" parent. If we find that our g-value is smaller then
		      ; our parent, we know we don't need to go anymore with this node.
		      (pop tree))
		  )
		 ( (evenp (depth-of parent-node)) ; is parent minizming?
		   (if (> Vx Vp)
		       (pop tree))) ; The same thing as above, but in this case
		 ; we are checking to see if is larger then our parents node.
		 )))
	     ); end of when
	   ); end of loop
	 (turn-board-around) ; We finished with our children (which req to be the 
	 ; the opponent), so lets put the right player back on track.

	 ; ----
	 ; Node with no children:
	 ; ----
	 (when (and (= (length (children-of first-node)) 0) (or (= *neg-infinity* Vx) (= *pos-infinity* Vx)))
		; in case we get a board where we cannot move.
		; we have to modify our parent to the most disadvantage number
		   (pop tree)
		   (setf (get parent-node 'g-value) 0)
		   ; this will direct the game to make the opponents table blank.
		   (cond
		    ( (oddp (depth-of parent-node))
		      (+ *pos-infinity* 1))
		    ( (evenp (depth-of parent-node))
		      (- *neg-infinity* 1))
		   ))

	 (if (not (= (depth-of first-node) 1)) ; if this is the top 'root' node
	     (setq tree (a-b-prunning tree)); if it is not, then we still need to process
	   (children-of first-node)) ; this is the top 'root' node. Returing its children,
	       ) ; end of condition
	    )
      )
    )
   )
  )



; +==========================================================================+
; | Conquering methods                                                       |
; |                                                                          |
; +==========================================================================+
;

; This methods does two things:
;
;  o  sows the rocks in next pits based on how many were taken (it does that in a circular fashion)
;  o  sows into the players bank (never into the opponents) whenever the bank is approached.
;  o  removes the rocks from the opponent based on a couple of rules:
;      o  if there are 2 or 3 rocks it takes them all
;      o  if there are 2 or 3 rocks on the opponents side on the counter-side, then they will taken
;         as well. 
;

(defun take-out-from-pit (node which-one)


  ; NOTE:
  ;
  ; When reading this, please take your time to read my comments. It is quite complicated.
  ;
  ; NOTE2:
  ;
  ; When I'm refering to "top" picture the top 4 pits (the opponent's/computer's). 
  ; When I'm refering to "bottom" picture the bottow 4 pits (mine/yours).
  ; So, if I start at the bottom, and sow my last stone in top, that means my last stone is in the 
  ; opponents area. Got it? :p

  (let
      (
       (new-board (copy-list (board-of node))) 
       ; throughout the method, we will only operate on this board
       (bank (copy-list (bank-of node)))              
       ; --------------------""----------------------------- bank
       (starting-index (+  (- which-one 1)  (current-player-starting-index-is))) ; our starting index in the list of pits.
					; the (- which-one 1) is used only so that one can invoke the method with a pit starting with 1, not zero.
       (stones-taken)                   ; how many stones we took from starting-index
       (index-of-end 0)                 ; where we finished sowing stones
       )
 
    (setq stones-taken (nth starting-index new-board)) ; take out all the stones to sow
    (setf (nth starting-index new-board) 0)    ; clears the pit

    ; -----
    ; Deposits stones
    ; -----

    (setq index-of-end (do*(
	 (index  starting-index  (mod   index   *amount-of-pits* )); our "marker" on each pit (which also wraps itself)
	 (deposited-stones 0 deposited-stones)              ; how many stones did we deposit
	 )
	((>= deposited-stones stones-taken) (mod  index  *amount-of-pits*))
					; ^- we return the location (index) of the last stone.

	(cond        					; check to see if we have to sow in our bank
	 ( (and (= *which-player* 0) (= (- (/ *amount-of-pits* 2) 1) index)) ; is the index a step in front of the BOTTOM bank's
					; and the current player is BOTTOM?
	   (incf (nth *which-player* bank) 1) ; yes, lets add a stone to the bank
	   (incf deposited-stones 1)
	   (unless (= (- deposited-stones stones-taken) 0) ; in case where we can add more stones
	     (incf index 1)
	     (incf (nth index new-board) 1)
	     (incf deposited-stones 1)
	     ; ^- deposit on stone in the next pit
	     ))
	 
	 ( (and (= *which-player* 1) (= (-  *amount-of-pits* 1) index))
	   ; is the index in front of TOP's bank and are we TOP?
	   (incf (nth *which-player* bank) 1) ; but of course, so lets add a pebble to our savings account :p
	   (incf deposited-stones 1)
	   (unless (= (- deposited-stones stones-taken) 0)	     
	     (setq index (mod (+ index 1) *amount-of-pits*)) 
	     (incf (nth index new-board) 1)
	     (incf deposited-stones 1)
	     
	     ))
	 (t
	  (setq index (mod (+ index 1) *amount-of-pits*)) ; Just adding the stones to each pit.
	  (incf deposited-stones 1)
	  (incf (nth index new-board) 1))

	 )
	))


    (incf index-of-end 1)  ; to make it easier for division of *amount-of-pits*. Look below.
    (incf starting-index 1) ; ----------------------""-------------------------------------

    ; ------
    ; Conquering part
    ; ------

    (cond

     ( (<= 1  index-of-end  (/ *amount-of-pits* 2)) 
       ; is the sowed stone in the "bottom" side (between index 0 and 3)

       (unless (<= 1  starting-index   (/ *amount-of-pits* 2)) 
        ; did we start at the "top" side? (between 3 and 7)

	; Yes, hence we can pick itmes clockwise (from left to right)	 	 
	 (do* (
	       (conquest-index  (- index-of-end 1)(mod (- conquest-index 1) (- *amount-of-pits* 1)))    ; the index-marker used for conquering stones. Also wraps around itself.
		; we are going toward right (clockwise)
	       )
		; exit condition:
		; is the element neither 2 or 3 and are we still at "bottom" side?
	     ((or (and (not (= (nth conquest-index new-board) 2)) (not (= (nth conquest-index new-board) 3))) (<= (/ (- *amount-of-pits* 1) 2)  conquest-index  *amount-of-pits* )) new-board)
		; ^- looks quite intimidating. Basicly the 'or' checks to see if the item is not 3 or not 2. The and checks previous statment and if we are trying to take items out of our own side.

	   (incf (nth (current-player-bank-index-is) bank) (nth conquest-index new-board)) ; yes, we are
   	   ; adding his stones to our bank! We are getting rich!
	   (setf (nth conquest-index new-board) 0)
	   )
	 )
       )

     ((<= (/ *amount-of-pits* 2) index-of-end *amount-of-pits*) 
      ; is the last stone on the "top" side (between 3 and 7)?

      (unless (<= (+ (/ *amount-of-pits* 2) 1)  starting-index *amount-of-pits*) 
        ; did we start at the "bottom" side ( between 0 and 3)?
	; Yes, our starting location is between 0-3 and we landed between 4-7, hence we can
	; go left counterclockwise (from right to left -- on the list)
	(do* (
	       (conquest-index (- index-of-end 1) (- conquest-index 1))    ; the index-marker used for conquering stones (no need t wrap around)
					; we are going toward left (counter-clockwise)
	       )
					; exit condition:
					; is the element neither 2 or 3 and are we still at "top" side?
	     ((or (and (not (= (nth conquest-index new-board) 2)) (not (= (nth conquest-index new-board) 3))) (<= 0 conquest-index (/ (- *amount-of-pits* 1) 2))) new-board)	  

	  (incf (nth (current-player-bank-index-is) bank) (nth conquest-index new-board)) ; yes, we are
					; adding his stones to our bank!
	  (setf (nth conquest-index new-board) 0)
	  )	 
	)
      )
     )
    ; whew.. returing our brand-new-spanky-node
    (if (equal new-board (board-of node)) ; did we accomplish anything?
	nil       ; no
					;
      (make-node new-board bank         ; yes, lets make a node with our new-board, bank and ...
		 (+ (depth-of node) 1)  ; since we are going one step depper down, we are adding our parent depth plus 1
					; Now this is the awfull part. We have to give -/+ infinity or a "correct" goodness value
					; depending on the depth of this node. If the new node is at depth *lookahead-depth* we will give
					; it a correct g-value, otherwise -/+ infinity depending whether the node will be maximized (odd depth)
					; or minimized (even depth)
		 
		 (cond
		  ( (= (+ (depth-of node) 1) *lookahead-depth*) ; is this a terminal node?		     
					; yes, hence
		    (- (nth 1 bank) (nth 0 bank))) ; we subtract the TOP player's bank from BOTTOM's player
		  ( (oddp (depth-of node))    ; was our parent node maxmized? (if so, then this node will be minimized		     
		    *pos-infinity*)         ; since this node is minimized, we pick positive infinity
		  ( (evenp (depth-of node))   ; vice-versa of the above.
		    *neg-infinity*)
		  )
		 node                     ; parent-node.
		 nil                      ; children-list
		 )
      )
    )
  )


(defun take-out-from-pit-1 (node)
  
  (take-out-from-pit node 1)
)


(defun take-out-from-pit-2 (node)
  
  (take-out-from-pit node 2)
)


(defun take-out-from-pit-3 (node)
  
  (take-out-from-pit node 3)
)


(defun take-out-from-pit-4 (node)
  
  (take-out-from-pit node 4)
)


; +==========================================================================+
; | Various important functions                                              |
; |                                                                          |
; +==========================================================================+
;

;
; Returns the index of the board for the _current_ player
; 
(defun current-player-starting-index-is ()

  ; We have to toggle a switch to be able to find out whoes turn it is

  (* *which-player*  (/ *amount-of-pits* 2))

)

(defun current-player-bank-index-is ()
  
   (* *which-player* 1)

)

;
; Turns the board around. 
;
(defun turn-board-around ()
  
  (if (= *which-player* 0) 
      (setq *which-player* 1)
    (setq *which-player* 0))

  )


;
; Creates the nodes for the list.
;

(defun make-node (board bank  depth g-value parent-node childrens)


  (let ((new-symbol (intern (symbol-name (gensym))))
	)
    (setf (get new-symbol 'board) board)
    (setf (get new-symbol 'bank) bank)
    (setf (get new-symbol 'g-value) g-value)
    (setf (get new-symbol 'depth) depth)
    (setf (get new-symbol 'parent-node) parent-node)
    (setf (get new-symbol 'childrens) childrens)
    new-symbol
    )
)

(defun board-of (node)
  
  (get node 'board))

(defun bank-of (node)
  
  (get node 'bank))

(defun g-value-of (node)

  (get node 'g-value))

(defun parent-of (node)

  (get node 'parent-node))

(defun depth-of (node)

  (get node 'depth))

(defun children-of (node)

  (get node 'childrens))


; +==========================================================================+
; | Messages                                                                 |
; |                                                                          |
; +==========================================================================+
;


(defun menu-game-message ()

  (format t "~%+-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=+~%")
  (format t "|                                                                  |~%")
  (format t "|  Welcome humanoid.                                               |~%")
  (format t "|                                                                  |~%")
  (format t "|  How do you wish to challenge the All Mighty HAL 9000?           |~%")
  (format t "|                                                                  |~%")
  (format t "|    [0]. No challenge (you play against another weak human)       |~%")
  (format t "|    [1]. You play against me and lose.                            |~%")
  (format t "|    [2]. I split in half and play against myself.                 |~%")
  (format t "|    [3]. Change difficulty level.                                 |~%")
  (format t "|    [4]. Hide away from me and quiver from fright.                |~%")
  (format t "|                                                                  |~%")
  (format t "+-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=+~%")
  (format t "~%Type either 0,1,2,3,4 and press Enter> ")
)



(defun show-pits-game-message (node)

  (let 
      ((L (board-of node))
       (bank (bank-of node))
       (player-status '(bottom top)))
       
  (format t "~%")
  (format t "+-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-+~%")
  (format t "|                                                         |~%")
  (format t "|   TOP      pit#4   pit#3   pit#2   pit#1                |~%")
  (format t "|   bank     +----+  +----+  +----+  +----+               |~%")
  (format t "|  +-----+   | ~A  |  | ~A  |  | ~A  |  | ~A  |    +-----+    |~%" (nth 7 L) (nth 6 L) (nth 5 L) (nth 4 L))
  (format t "|  |     |   +----+  +----+  +----+  +----+    |     |    |~%")
  (format t "|  |  ~A  |                                     |  ~A  |    |~%" (nth 1 bank) (nth 0 bank))
  (format t "|  |     |                                     |     |    |~%")
  (format t "|  +-----+   +----+  +----+  +----+  +-----+   +-----+    |~%")
  (format t "|            | ~A  |  | ~A  |  | ~A  |  |  ~A  |    BOTTOM    |~%" (nth 0 L) (nth 1 L) (nth 2 L) (nth 3 L))
  (format t "|            +----+  +----+  +----+  +-----+    bank      |~%")
  (format t "|            pit#1   pit#2   pit#3   pit#4                |~%")
  (format t "|                                                         |~%")
  (format t "+-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-+~%")
  (format t "~%You own the ~A part of the board~%" (nth *which-player* player-status))
  (format t "~%Please choose the pit from which you want to scop the stones> ")
))


(defun lost-game-message ()

  (let
      ((player '(-bottom- ---top--)))
  (format t"~%+-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-+~%")
  (format t"|                               |~%")
  (format t"| ~A player is a LOSER!   |~%" (nth *which-player* player))
  (format t"|                               |~%")
  (format t"+-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-+~%"))
)


(defun won-game-message ()

  (format t"~%+-=-=-=-=-=-=-=-=-=-=-=-+~%")
  (format t"|                       |~%")
  (format t"| You  won the game!    |~%" )
  (format t"|                       |~%")
  (format t"+-=-=-=-=-=-=-=-=-=-=-=-+~%~%")
)

(defun tie-game-message ()
  (princ *which-player*)
 (format t"~%+-=-=-=-=-=-=-=-+~%")
  (format t"|               |~%")
  (format t"|   TIE!        |~%")
  (format t"|               |~%")
  (format t"+-=-=-=-=-=-=-=-+~%")
)
